import pytest

from briefly.model.style import NotionStyle
from briefly.rendering.graphs import build_pie_chart_bytes
from briefly.rendering.pdf_generator import PDF


@pytest.fixture
def pdf() -> PDF:
    pdf = PDF(NotionStyle())
    pdf.add_page()
    return pdf


@pytest.fixture
def data() -> dict[str, float]:
    return {
        "one": 12,
        "two": 0,
        "three": 30,
        "four": 55,
    }


def test_graph_with_no_data_returns_none():
    assert build_pie_chart_bytes([0, 0, 0]) is None


def test_graph():
    assert build_pie_chart_bytes([1, 2, 3]) is not None


def test_graph_with_negative_values():
    with pytest.raises(ValueError):
        build_pie_chart_bytes([-1, -2, -3])


def test_header(pdf: PDF):
    pdf.main_title("TEST - Header")
    assert pdf.font_family == "inter"
    assert pdf.get_y() == 55
    assert pdf.get_x() == 25


def test_section_title(pdf: PDF):
    pdf.section_title("Test section")
    assert pdf.font_family == "inter"
    assert pdf.font_size_pt == 13
    assert pdf.get_y() == 45
    assert pdf.get_x() == 25


def test_summary_card(pdf: PDF):
    (x, y) = pdf.summary_card(["Test summary card"])
    assert pdf.font_family == "inter"
    assert pdf.font_size_pt == 10
    assert x == 105
    assert y == 25 + 6 + 10


def test_styled_table(pdf: PDF):
    pdf.styled_table(
        ["header1", "header2", "header3", "header4"],
        [
            ["asd", "asd", "asd", "asd"],
            ["asd", "asd", "asd", "asd"],
            ["asd", "asd", "asd", "asd"],
        ],
        [30, 15, 20, 5],
    )
    assert pdf.font_family == "inter"
    assert pdf.font_size_pt == 7
    assert pdf.get_x() == 25
    assert pdf.get_y() == 66


def test_tag(pdf: PDF):
    width, height = pdf.tag("Test tag")
    assert pdf.font_family == "inter"
    assert pdf.font_size_pt == 7
    assert width == pytest.approx(13.15, 0.01)
    assert height == 5


def test_pie_chart(pdf: PDF, data: dict[str, float]):
    x, y = pdf.pie_chart(data, "Test Pie Chart")
    assert pdf.font_family == "inter"
    assert pdf.font_size_pt == 10
    assert x == 123
    assert y == 54



import pytest

from briefly.model.style import MochaStyle
from briefly.rendering.pdf_generator import PDF

DARK_BACKGROUND = (38, 33, 43)


@pytest.mark.skip(reason="Manually run")
def test():
    style = MochaStyle()
    pdf = PDF(style)
    pdf.add_page()

    pdf.main_title("JIRA Summary Test")

    pdf.section_title("Overview")
    pdf.summary_card(
        ["Total Tickets: 58", "Completed: 42", "In Progress: 10", "Blocked: 6"],
        width=50,
    )

    pdf.summary_card(
        ["Total Tickets: 58", "Completed: 42", "In Progress: 10", "Blocked: 6"],
        width=50,
    )

    (_, y) = pdf.summary_card(
        ["Total Tickets: 58", "Completed: 42", "In Progress: 10", "Blocked: 6"],
        width=50,
    )

    pdf.set_y(y + 10)

    pdf.styled_table(
        headers=["Key", "Summary", "Status", "Assignee"],
        rows=[
            ["PROJ-101", "Fix login flow", "Done", "Alice"],
            ["PROJ-102", "Add metrics dashboard", "In Progress", "Bob"],
            ["PROJ-103", "Payment gateway issue", "Blocked", "Eve"],
        ],
        col_widths=[20, 80, 30, 30],
    )

    data = {
        "cat1": 10,
        "cat3": 8,
        "cat2": 5,
        "cat4": 3,
        "cat5": 2,
        "cat6": 20,
        "cat7": 15,
        "cat8": 10,
    }

    pdf.bar_chart(data, caption="Categories", height=30)

    pdf.pie_chart(
        data,
        height=30,
        caption="Categories",
    )

    pdf.pie_chart(
        data,
        height=30,
        caption="Categories",
    )

    pdf.bar_chart(data, caption="Categories", height=30)
    pdf.bar_chart(data, caption="Categories", height=30)
    pdf.output("./output/test.pdf")



import subprocess


def lint() -> None:
    subprocess.run(["ruff", "check", ".", "--fix"], check=True)


def format() -> None:
    subprocess.run(["ruff", "format", "."], check=True)


def typecheck() -> None:
    subprocess.run(["mypy", "src"], check=True)


def coverage() -> None:
    subprocess.run(["pytest", "--cov-report=html"])



from typing import Tuple

STRIPE_DARK = (10, 37, 64)
STRIPE_LIGHT_BG = (246, 249, 252)
STRIPE_GRAY_TEXT = (66, 84, 102)
STRIPE_HEADER_BG = (230, 236, 244)
STRIPE_ACCENT = (99, 91, 255)
STRIPE_ROW_ALT = (242, 245, 252)

MINIMALIST_HEADER_BG = (45, 62, 80)
MINIMALIST_TEXT_COLOR = (255, 255, 255)
MINIMALIST_TEXT_COLOR_GRAY = (45, 62, 80)
MINIMALIST_ROW_ALT = (247, 249, 252)


class Style:
    background_color: Tuple[int, int, int]
    chart_colors: list[Tuple[int, int, int]]
    priority_color: Tuple[int, int, int]
    card_background: Tuple[int, int, int]
    header_background: Tuple[int, int, int]
    table_header_color: Tuple[int, int, int]
    table_row_colors: list[Tuple[int, int, int]]
    font_color: Tuple[int, int, int]
    section_title_color: Tuple[int, int, int]
    card_details_color: Tuple[int, int, int]
    border_color: Tuple[int, int, int] = (230, 230, 230)
    header_color: Tuple[int, int, int] = (55, 53, 47)
    disabled_color: Tuple[int, int, int]


class PurpleHaze(Style):
    background_color = (255, 255, 255)
    font_color = (43, 34, 43)
    header_color = (255, 255, 255)
    header_background = (129, 126, 160)
    border_color = (232, 227, 232)
    priority_color = (191, 106, 95)
    disabled_color = (167, 154, 167)
    chart_colors = [
        (68, 68, 118),
        (91, 91, 135),
        (115, 115, 152),
        (138, 138, 169),
        (161, 161, 186),
        (185, 185, 203),
        (208, 208, 220),
        (232, 232, 238),
    ]


class NotionStyle(Style):
    priority_color: Tuple[int, int, int] = (252, 216, 212)

    chart_colors: list[Tuple[int, int, int]] = [
        (155, 207, 87),  # green
        (246, 199, 68),  # yellow
        (108, 155, 245),  # blue
        (221, 148, 255),  # purple
        (255, 170, 153),  # coral
        (181, 181, 181),  # gray
        (212, 228, 247),
        (255, 232, 163),
        (252, 216, 212),
        (217, 241, 208),
    ]
    card_background: Tuple[int, int, int] = (255, 238, 189)
    header_background: Tuple[int, int, int] = (247, 246, 243)
    table_header_color: Tuple[int, int, int] = (243, 242, 239)
    table_row_colors: list[Tuple[int, int, int]] = [(255, 255, 255), (250, 249, 247)]
    font_color: Tuple[int, int, int] = (55, 53, 47)
    section_title_color: Tuple[int, int, int] = (55, 53, 47)
    card_details_color: Tuple[int, int, int] = (80, 79, 75)
    disabled_color: Tuple[int, int, int] = (165, 159, 141)
    background_color: Tuple[int, int, int] = (255, 255, 255)


class LatteStyle(Style):
    background_color: Tuple[int, int, int] = (239, 241, 245)
    chart_colors: list[Tuple[int, int, int]] = [
        (220, 138, 120),
        (221, 120, 120),
        (234, 118, 203),
        (136, 57, 239),
        (210, 15, 57),
        (230, 69, 83),
        (254, 100, 11),
        (223, 142, 29),
        (64, 160, 43),
        (23, 146, 153),
        (4, 165, 229),
        (32, 159, 181),
        (30, 102, 245),
        (114, 135, 253),
    ]
    priority_color: Tuple[int, int, int] = (210, 15, 57)
    card_background: Tuple[int, int, int] = (220, 224, 232)
    header_background: Tuple[int, int, int] = (220, 138, 120)
    table_header_color: Tuple[int, int, int] = (172, 176, 190)
    table_row_colors: list[Tuple[int, int, int]] = [(220, 224, 232), (188, 192, 204)]
    font_color: Tuple[int, int, int] = (76, 79, 105)
    section_title_color: Tuple[int, int, int] = (92, 95, 119)
    card_details_color: Tuple[int, int, int] = (92, 95, 119)
    border_color: Tuple[int, int, int] = (156, 160, 176)
    header_color = (76, 79, 105)
    disabled_color = (188, 192, 204)


class MochaStyle(Style):
    background_color: Tuple[int, int, int] = (30, 30, 46)
    chart_colors: list[Tuple[int, int, int]] = [
        (245, 224, 220),
        (242, 205, 205),
        (245, 194, 231),
        (203, 166, 247),
        (243, 139, 168),
        (235, 160, 172),
        (250, 179, 135),
        (249, 226, 175),
        (166, 227, 161),
        (148, 226, 213),
        (137, 220, 235),
        (116, 199, 236),
        (137, 180, 250),
        (180, 190, 254),
    ]
    priority_color: Tuple[int, int, int] = (249, 226, 175)
    card_background: Tuple[int, int, int] = (24, 24, 37)
    header_background: Tuple[int, int, int] = (49, 50, 68)
    table_header_color: Tuple[int, int, int] = (69, 71, 90)
    table_row_colors: list[Tuple[int, int, int]] = [(30, 30, 46), (24, 24, 37)]
    font_color: Tuple[int, int, int] = (205, 214, 244)
    section_title_color: Tuple[int, int, int] = (186, 194, 222)
    card_details_color: Tuple[int, int, int] = (186, 194, 222)
    border_color: Tuple[int, int, int] = (49, 50, 68)
    header_color = (205, 214, 244)
    disabled_color = (73, 77, 100)



from io import BytesIO
from typing import Optional

from matplotlib import pyplot as plt

NOTION_CHART_COLORS = [
    (155, 207, 87),  # green
    (246, 199, 68),  # yellow
    (108, 155, 245),  # blue
    (221, 148, 255),  # purple
    (255, 170, 153),  # coral
    (181, 181, 181),  # gray
]


def build_pie_chart_bytes(
    values: list[float],
    size: float = 35,
    colors: Optional[list[tuple[int, int, int]]] = None,
) -> Optional[BytesIO]:
    """
    Return a PNG image as bytes for a pie chart.
    :param values: The values to plot
    :param size: The size of the chart in mm
    :param colors: Optional list of colors to use for each value
    :return: the bytes of the chart or None if there are no values
    """

    size_inch = size / 25.4

    if sum(values) == 0:
        return None

    fig, ax = plt.subplots(figsize=(size_inch, size_inch))
    colors = colors or NOTION_CHART_COLORS
    graph_colors = [(r / 255, g / 255, b / 255) for r, g, b in colors[: len(values)]]
    ax.pie(values, colors=graph_colors, startangle=90, counterclock=False)

    buf = BytesIO()
    plt.tight_layout(pad=0)
    plt.savefig(buf, format="png", dpi=200, transparent=True)
    plt.close(fig)
    buf.seek(0)
    return buf



from dataclasses import dataclass
from typing import Literal

FONT_FAMILY: str = "Inter"
ICON_FONT_FAMILY: str = "Material Icons"


@dataclass(frozen=True)
class FontSpec:
    family: str
    style: Literal["", "B", "I"]
    filename: str


FONTS: tuple[FontSpec, ...] = (
    FontSpec(FONT_FAMILY, "", "Inter-Regular.ttf"),
    FontSpec(FONT_FAMILY, "B", "Inter-Bold.ttf"),
    FontSpec(FONT_FAMILY, "I", "Inter-Italic.ttf"),
    FontSpec(ICON_FONT_FAMILY, "", "MaterialIcons-Regular.ttf"),
)



from datetime import datetime
from importlib.resources import files
from typing import Any, List, Optional, Tuple

from fpdf import FPDF, YPos

from briefly.model.style import Style, PurpleHaze
from briefly.rendering.font_spec import FONT_FAMILY, FONTS, ICON_FONT_FAMILY
from briefly.rendering.graphs import build_pie_chart_bytes
from briefly.rendering.icons import FLAG_ICON

HEADER_SIZE: int = 20
SECTION_TITLE_SIZE: int = 13
TEXT_SIZE: int = 10
ICON_FONT_SIZE: int = 7
LABEL_SIZE: int = 7
MARGIN_SIZE: int = 25

_SMALL_SPACING: float = 2
_MEDIUM_SPACING: float = 5
_LARGE_SPACING: float = 10


class PDF(FPDF):
    style: Style

    def __init__(self, style: Style = PurpleHaze(), **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.style = style
        self.set_margin(MARGIN_SIZE)
        self.set_page_background(style.background_color)
        self.generation_time = datetime.now()
        self.setup_fonts()

    def setup_fonts(self) -> None:
        font_pkg = files("briefly.fonts")
        for font in FONTS:
            self.add_font(font.family, font.style, str(font_pkg / font.filename))
        self.set_font(FONT_FAMILY, "", TEXT_SIZE)

    def footer(self) -> None:
        self.set_y(-15)
        self.set_font(FONT_FAMILY, "I", 7)
        time_str: str = self.generation_time.strftime("%d.%m.%Y %H:%M:%S")
        self.cell(0, 10, time_str, align="L")
        self.cell(0, 10, f"Page {self.page_no()}", align="R")

    def main_title(self, text: str) -> None:
        self.set_font(FONT_FAMILY, "B", size=HEADER_SIZE)
        self.set_fill_color(*self.style.header_background)
        self.set_text_color(*self.style.header_color)
        self.cell(0, HEADER_SIZE, text, align="C", fill=True, new_y=YPos.NEXT)
        self.set_text_color(*self.style.font_color)
        self.set_y(self.get_y() + _LARGE_SPACING)

    def divider(self) -> None:
        self.set_draw_color(*self.style.border_color)
        x1, x2 = MARGIN_SIZE, self.w - MARGIN_SIZE
        y = self.get_y() + _MEDIUM_SPACING
        self.line(x1, y, x2, y)
        self.ln(_MEDIUM_SPACING)

    def section_title(self, text: str, link: Optional[str] = None) -> None:
        self.set_y(self.get_y() + _MEDIUM_SPACING)
        self._break_page_if_needed(content_height=40)
        self.set_font(FONT_FAMILY, "B", SECTION_TITLE_SIZE)
        self.set_text_color(*self.style.section_title_color)
        self.cell(0, 10, text, link=link or 0, new_y=YPos.NEXT)
        self.set_y(self.get_y() + _MEDIUM_SPACING)

    def summary_card(self, items: List[str], width: int = 80) -> tuple[float, float]:
        padding = _MEDIUM_SPACING
        row_height = 6
        card_height = (len(items) * row_height) + 2 * padding
        self._break_page_if_needed(card_height)
        start_x, start_y = self.x, self.y

        if start_y + card_height >= self.h - self.b_margin:
            self.add_page()

        self.set_fill_color(*self.style.card_background)
        self.rect(
            start_x,
            start_y,
            width,
            card_height,
            style="F",
            round_corners=True,
            corner_radius=1.5,
        )
        self.set_font(FONT_FAMILY, "", TEXT_SIZE)
        self.set_text_color(*self.style.font_color)

        x = start_x + padding
        y = start_y + padding
        self.set_text_color(*self.style.card_details_color)
        for text in items:
            self.set_xy(x, y)
            self.cell(width - 2 * padding, row_height, text, align="L")
            y = y + row_height

        if start_x + width >= self.w - self.r_margin:
            self.set_xy(self.l_margin, start_y + card_height + padding)
        else:
            self.set_xy(start_x + width + padding, start_y)
        return start_x + width, start_y + card_height

    def styled_table(
        self,
        headers: list[str],
        rows: list[list[str]],
        col_widths: list[int],
    ) -> None:
        self.set_font(FONT_FAMILY, "B", TEXT_SIZE)
        self.set_fill_color(*self.style.table_header_color)
        self.set_text_color(*self.style.font_color)

        for i, h in enumerate(headers):
            self.cell(col_widths[i], 10, h, border="B", fill=True)
        self.ln(10)

        # rows
        self.set_font(FONT_FAMILY, "", LABEL_SIZE)

        for idx, row in enumerate(rows):
            self.set_fill_color(*self.style.table_row_colors[idx % 2])

            for i, cell in enumerate(row):
                self.cell(col_widths[i], LABEL_SIZE, cell, border="B", fill=True)
            self.set_fill_color(*self.style.card_background)
            self.ln()

        self.set_fill_color(*self.style.card_background)
        self.set_y(self.get_y() + _LARGE_SPACING)

    def tag(self, text: str) -> Tuple[float, float]:
        bg = self.style.background_color
        self.set_font(FONT_FAMILY, "", LABEL_SIZE)
        self.set_text_color(*self.style.font_color)

        text_w = self.get_string_width(text) + _SMALL_SPACING * 2
        text_h = 5
        x, y = self.x, self.y

        self.set_fill_color(*bg)
        self.rect(
            x, y, text_w, text_h, style="F", round_corners=True, corner_radius=1.5
        )

        self.cell(text_w, text_h, text, align="C")
        self.set_x(x + text_w)
        return text_w, text_h

    # def detailed_tickets_table(self, tickets: list[Ticket]) -> None:
    #     self.set_font(FONT_FAMILY, "", TEXT_SIZE)
    #
    #     for t in tickets:
    #         self.ticket_card_long(t)

    # def ticket_card_long(self, ticket: Ticket) -> None:
    #     width = self.w - self.r_margin - self.l_margin
    #     height = 16
    #     self.__break_page_if_needed(height)
    #     start_x = self.x
    #     start_y = self.y
    #     left_padding = 6
    #     block_width = 20
    #
    #     stripe_color: tuple[int, int, int] = self.style.category_colors.get(
    #         ticket.category, self.style.border_color
    #     )
    #     self.accent_card(stripe_color, width, height)
    #
    #     self.set_xy(start_x + left_padding, start_y + _SMALL_SPACING)
    #     key_width = 12
    #     self.tag(ticket.status.value)
    #     self.set_font(FONT_FAMILY, "B", LABEL_SIZE)
    #     self.set_text_color(*self.style.font_color)
    #     self.set_x(start_x + left_padding + block_width)
    #     self.cell(key_width, 5, ticket.key, align="R")
    #     self.set_font(FONT_FAMILY, "", TEXT_SIZE)
    #     title_width = (
    #             width - key_width - left_padding - _SMALL_SPACING - block_width - 5
    #     )
    #     self.cell(title_width, 5, ticket.summary)
    #     if ticket.flagged:
    #         self._flagged_icon(self.x + 2, self.y)
    #
    #     self.set_xy(start_x + left_padding, self.y + 5 + _SMALL_SPACING)
    #     if ticket.priority:
    #         dot_color: tuple[int, int, int] = self.style.priority_colors.get(
    #             ticket.priority, (217, 241, 208)
    #         )
    #         self.legend_label(dot_color, ticket.priority)
    #
    #     self.set_x(start_x + left_padding + block_width)
    #     self.tag(ticket.issue_type)
    #     self.set_x(start_x + left_padding + block_width * 2)
    #     self.tag(f"SP: {ticket.story_points or 'N/A'}")
    #     self.set_x(start_x + left_padding + block_width * 3)
    #     if ticket.component:
    #         self.tag(ticket.component)
    #     self.set_y(start_y + height + _MEDIUM_SPACING)

    def _break_page_if_needed(self, content_height: float) -> None:
        if self.y + content_height >= self.h - self.b_margin:
            self.add_page()

    # def ticket_card_short(self, ticket: Ticket) -> None:
    #     width = 77.5
    #     height = 30
    #     self.__break_page_if_needed(height)
    #
    #     start_x, start_y = self.x, self.y
    #     row_height = 5
    #
    #     stripe_color: tuple[int, int, int] = self.style.category_colors.get(
    #         ticket.category, self.style.border_color
    #     )
    #     self.accent_card(stripe_color, width, height)
    #
    #     if ticket.flagged:
    #         self.set_text_color(*self.style.disabled_color)
    #
    #     text_start_x = start_x + 12.5
    #     self.set_font(FONT_FAMILY, "B", LABEL_SIZE)
    #     self.set_xy(text_start_x, start_y + 9)
    #     self.cell(
    #         15, row_height, ticket.key, align="R", new_x=XPos.LEFT, new_y=YPos.NEXT
    #     )
    #     _, y = self._two_line_label(ticket.status.value, text_start_x, self.y + 1)
    #     y = y + 1
    #     self.set_xy(text_start_x - 3, y + 0.2)
    #     self.set_font(ICON_FONT_FAMILY, "", 7)
    #     self.cell(3, row_height, DUE_DATE_ICON, align="L")
    #     x, _ = self._small_label(
    #         ticket.due_date.strftime("%d.%m.%Y") if ticket.due_date else "",
    #         text_start_x,
    #         y,
    #     )
    #
    #     x, _ = self._small_label(ticket.priority or "N/A", x, y)
    #     story_points_text = f"SP: {ticket.story_points or 'N/A'}"
    #     x, _ = self._small_label(story_points_text, x + 6, y)
    #     if ticket.flagged:
    #         self._flagged_icon(x + 8, y)
    #
    #     self.set_font(FONT_FAMILY, "", TEXT_SIZE)
    #     self.set_xy(text_start_x + 15, start_y + 4)
    #     summary_width = self.get_string_width(ticket.summary)
    #     if summary_width <= 45:
    #         self.cell(45, 4, ticket.summary, align="L")
    #     elif summary_width >= 150:
    #         self.multi_cell(
    #             45, 15, f"{ticket.summary[:70]}...", max_line_height=4, align="L"
    #         )
    #     else:
    #         self.multi_cell(45, 14, ticket.summary, max_line_height=4, align="L")
    #
    #     self.set_text_color(*self.style.font_color)
    #     if start_x == self.l_margin:
    #         self.set_xy(start_x + width + _MEDIUM_SPACING, start_y)
    #     else:
    #         self.set_y(start_y + height + _MEDIUM_SPACING)

    def _small_label(self, text: str, x: float, y: float) -> tuple[float, float]:
        self.set_font(FONT_FAMILY, "", LABEL_SIZE)
        self.set_xy(x, y)
        self.cell(15, 5, text, align="R")
        return x + 15, self.y + 3

    def _two_line_label(self, text: str, x: float, y: float) -> tuple[float, float]:
        self.set_font(FONT_FAMILY, "", LABEL_SIZE)
        self.set_xy(x, y)
        self.multi_cell(15, 6, text, align="R", max_line_height=3)

        return x + 15, self.y

    def _flagged_icon(self, x: float, y: float) -> None:
        self.set_font(ICON_FONT_FAMILY, "", ICON_FONT_SIZE)
        self.set_text_color(*self.style.priority_color)
        self.set_xy(x, y + 0.3)
        self.cell(3, 5, FLAG_ICON, align="R")
        self.set_xy(self.x, y - 0.3)
        self.set_text_color(*self.style.disabled_color)

    def _plot_bar_chart(
        self, values: list[float], height: float
    ) -> tuple[float, float]:
        spacing = 2
        bar_width = 3
        x = self.x
        start_y = self.y
        width = x + len(values) * (bar_width + spacing) + spacing
        max_value = max(values) if values else 0

        self.set_draw_color(*self.style.border_color)
        self.line(x, start_y, width, start_y)
        self.line(x, start_y + 5, width, start_y + 5)
        self.line(x, start_y + 10, width, start_y + 10)
        self.line(x, start_y + 15, width, start_y + 15)
        self.line(x, start_y + 20, width, start_y + 20)
        self.line(x, start_y + 25, width, start_y + 25)

        x += spacing

        for index, value in enumerate(values):
            if value > 0:
                self.set_fill_color(
                    *self.style.chart_colors[index % len(self.style.chart_colors)]
                )
                bar_height = height * value / max_value
                y = start_y + height - bar_height
                if bar_height > 1.5:
                    self.rect(
                        x,
                        y,
                        bar_width,
                        bar_height,
                        style="F",
                        round_corners=True,
                        corner_radius=1.5,
                    )
            x += bar_width + spacing

        return x - spacing, start_y + height

    def bar_chart(
        self, data: dict[str, float], caption: str, height: float
    ) -> tuple[float, float]:
        self._break_page_if_needed(height)
        start_x, start_y = self.x, self.y
        x, y = self._plot_bar_chart(list(data.values()), height)
        if len(data.keys()) > 12:
            x, y = start_x, y + _SMALL_SPACING
        else:
            x, y = x + _SMALL_SPACING, start_y + _SMALL_SPACING

        legend_labels = [f"{key} ({data[key]:.2f})" for key in data.keys()]
        x, y = self.legend(legend_labels, x, y, caption)
        end_x = x
        if start_x == self.l_margin:
            self.set_xy(end_x + _LARGE_SPACING, start_y)
        else:
            self.set_xy(self.l_margin, y + _LARGE_SPACING)
        return end_x, y

    def pie_chart(
        self,
        data: dict[str, float],
        caption: str,
        height: float = 70,
    ) -> tuple[float, float]:
        """Generate a pie chart in-memory and insert it into the PDF."""
        self._break_page_if_needed(height)

        img_buf = build_pie_chart_bytes(
            list(data.values()), colors=self.style.chart_colors
        )

        if img_buf is None:
            return self.x, self.y

        start_x = self.x
        x, y = self.x, self.y
        self.image(img_buf, x=x, y=y, w=height)
        self.set_xy(x + height, y)

        if len(data.keys()) > 12:
            legend_x, legend_y = start_x, y + height + _SMALL_SPACING
        else:
            legend_x = x + height + _MEDIUM_SPACING
            legend_y = y + _SMALL_SPACING
        legend_labels = [f"{key} ({data[key]})" for key in data.keys()]
        end_x, end_y = self.legend(legend_labels, legend_x, legend_y, caption)
        if x == self.l_margin:
            self.set_xy(end_x + _LARGE_SPACING, y)
        else:
            self.set_xy(self.l_margin, y + height + _LARGE_SPACING)
        return end_x, end_y

    def legend(
        self, labels: list[str], x: float, y: float, caption: str
    ) -> tuple[float, float]:
        self.set_xy(x, y)
        self.set_font(FONT_FAMILY, "", 9)
        self.cell(0, 5, caption, align="L", new_y=YPos.NEXT)
        legend_start_y = self.y + _SMALL_SPACING
        next_column_x = x + _SMALL_SPACING
        max_y = y

        legend_colors = self.style.chart_colors
        for idx, label in enumerate(labels):
            if idx % 4 == 0:
                self.set_xy(next_column_x, legend_start_y)
            color = legend_colors[idx % len(legend_colors)]
            x, y = self.legend_label(color, label)
            next_column_x = max(next_column_x, x + _LARGE_SPACING)

            max_y = max(max_y, y)

        return next_column_x, max_y

    def legend_label(
        self, color: tuple[int, int, int], label: str
    ) -> tuple[float, float]:
        """
        Generates a legend label with a colored dot. The label object has the height of 5mm.
        :param color: the color of the dot
        :param label: the text of the label
        :return: the position of the bottom right corner of the label
        """
        start_x, start_y = self.x, self.y
        self.set_fill_color(*color)
        self.ellipse(start_x + 1, start_y + 1.5, 2, 2, style="F")
        self.set_x(start_x + 3)
        self.set_font(FONT_FAMILY, size=LABEL_SIZE)
        self.set_text_color(*self.style.font_color)
        text_length = self.get_string_width(label)
        self.cell(15, 5, label, new_y=YPos.NEXT)
        self.set_font(FONT_FAMILY, size=TEXT_SIZE)
        self.set_x(start_x)
        return start_x + text_length, self.y

    def accent_card(
        self, accent_color: tuple[int, int, int], width: float, height: float
    ) -> None:
        self.set_draw_color(*self.style.border_color)
        self.rect(
            self.x,
            self.y,
            width,
            height,
            style="D",
            round_corners=True,
            corner_radius=2,
        )

        self.set_fill_color(*accent_color)
        self.rect(
            self.x,
            self.y,
            2,
            height,
            style="F",
            round_corners=True,
            corner_radius=2.2,
        )

    def bar_chart_with_limit(
        self, data: dict[str, float], limit: float, caption: str, height: float
    ) -> tuple[float, float]:
        self._break_page_if_needed(height)
        start_x, start_y = self.x, self.y
        x, y = self._plot_bar_chart_with_limit(list(data.values()), height, limit)
        if len(data.keys()) > 12:
            x, y = start_x, y + _SMALL_SPACING
        else:
            x, y = x + _SMALL_SPACING, start_y + _SMALL_SPACING

        legend_labels = [f"{key} ({data[key]:.2f})" for key in data.keys()]
        x, y = self.legend(legend_labels, x, y, caption)
        end_x = self.x
        if start_x == self.l_margin:
            self.set_xy(self.x + _LARGE_SPACING, start_y)
        else:
            self.set_xy(self.l_margin, y + _LARGE_SPACING)
        return end_x, y

    def _plot_bar_chart_with_limit(
        self, values: list[float], height: float, limit: float
    ) -> tuple[float, float]:
        spacing = 2
        bar_width = 3
        start_x = self.x
        start_y = self.y
        max_value = max(max(values), limit) if values else 0
        x = start_x + spacing

        for index, value in enumerate(values):
            if value > 0:
                self.set_fill_color(
                    *self.style.chart_colors[index % len(self.style.chart_colors)]
                )
                bar_height = height * value / max_value
                y = start_y + height - bar_height
                if bar_height > 1.5:
                    self.rect(
                        x,
                        y,
                        bar_width,
                        bar_height,
                        style="F",
                        round_corners=True,
                        corner_radius=1.5,
                    )
            x += bar_width + spacing

        limit_line_y = start_y + height - height * limit / max_value
        self.set_draw_color(*self.style.priority_color)
        self.set_line_width(0.4)
        self.line(start_x, limit_line_y, x - spacing, limit_line_y)
        self.set_line_width(0.2)
        return x - spacing, start_y + height
